#include "main.h"

void rendermon(struct wl_listener *listener, void *data) {
	struct output *output = wl_container_of(listener, output, frame);
	struct server *server = output->server;
	struct wlr_output *wlr_output = data;

	struct timespec now;
	clock_gettime(CLOCK_MONOTONIC, &now);
	output->last_frame = now;

	int width, height;
	wlr_output_effective_resolution(output->wlr_output, &width, &height);
	wlr_renderer_begin(server->renderer, width, height);

	float color[4] = {0.3, 0.3, 0.3, 1.0};
	wlr_renderer_clear(server->renderer, color);

	wlr_renderer_end(server->renderer);
	wlr_output_commit(output->wlr_output);
}

void cleanupmon(struct wl_listener *listener, void *data) {
	struct output *output = wl_container_of(listener, output, destroy);

	wl_list_remove(&output->link);
	wl_list_remove(&output->destroy.link);
	wl_list_remove(&output->frame.link);

	free(output);
}

void createmon(struct wl_listener *listener, void *data) {
	struct server *server = wl_container_of(listener, server, new_output);
	struct wlr_output *wlr_output = data;

	struct output *output = malloc(sizeof(struct output));
	if (output == NULL) {
		wlr_backend_destroy(server->backend);
		wl_display_destroy(server->wl_display);
		die("createmon", 1);
	}

	if (wl_list_length(&wlr_output->modes) > 0) {
		struct wlr_output_mode *mode = wl_container_of((&wlr_output->modes)->prev, mode, link);
		wlr_output_set_mode(wlr_output, mode);
	}

	clock_gettime(CLOCK_MONOTONIC, &output->last_frame);

	output->server = server;
	output->wlr_output = wlr_output;
	wl_list_insert(&server->outputs, &output->link);

	output->destroy.notify = cleanupmon;
	wl_signal_add(&wlr_output->events.destroy, &output->destroy);

	output->frame.notify = rendermon;
	wl_signal_add(&wlr_output->events.frame, &output->frame);

	wlr_output_schedule_frame(wlr_output);
}

void server_init(struct server *server) {
	server->wl_display = wl_display_create();
	assert(server->wl_display);

	server->wl_event_loop = wl_display_get_event_loop(server->wl_display);
	assert(server->wl_event_loop);

	server->backend = wlr_backend_autocreate(server->wl_display);
	assert(server->backend);

	server->renderer = wlr_renderer_autocreate(server->backend);
	assert(server->renderer);

	wlr_data_device_manager_create(server->wl_display);
	wlr_export_dmabuf_manager_v1_create(server->wl_display);
	wlr_screencopy_manager_v1_create(server->wl_display);
	wlr_data_control_manager_v1_create(server->wl_display);
	wlr_gamma_control_manager_v1_create(server->wl_display);
	wlr_primary_selection_v1_device_manager_create(server->wl_display);

	/* Init monitors */
	wl_list_init(&server->outputs);
	server->new_output.notify = createmon;
	wl_signal_add(&server->backend->events.new_output, &server->new_output);

	const char *socket = wl_display_add_socket_auto(server->wl_display);
	assert(socket);

	printf("Running compositor on wayland display '%s'\n", socket);
	setenv("WAYLAND_DISPLAY", socket, 1);

	if (!wlr_backend_start(server->backend)) {
		wl_display_destroy(server->wl_display);
		die("wlr_backend_start", 1);
	}
}

int main(void) {
	if (!getenv("XDG_RUNTIME_DIR"))
		die("XDG_RUNTIME_DIR must be set", 0);

	int sig[] = {SIGCHLD, SIGINT, SIGTERM, SIGPIPE};
	for (size_t i = 0; i < sizeof(sig) / sizeof(sig[0]); i++)
			signal(sig[i], NULL);

	wlr_log_init(LOG_LEVEL, NULL);

	struct server server;
	server_init(&server);

	wl_display_run(server.wl_display);
	wlr_backend_destroy(server.backend);
	wl_display_destroy(server.wl_display);
	return 0;
}
